(* Standard CMaps *)

(* Copyright 1990-2019 Adobe. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

Neither the name of Adobe nor the names of its contributors may be used to
endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. *)

let cmap_90ms_rksj_ucs2 () = __DATA:90ms-RKSJ-UCS2

let cmap_90pv_rksj_ucs2 () = __DATA:90pv-RKSJ-UCS2

let cmap_90pv_rksj_ucs2c () = __DATA:90pv-RKSJ-UCS2C

let cmap_adobe_cns1_b5pc () = __DATA:Adobe-CNS1-B5pc

let cmap_adobe_cns1_eten_B5 () = __DATA:Adobe-CNS1-ETen-B5

let cmap_adobe_cns1_h_cid () = __DATA:Adobe-CNS1-H-CID

let cmap_adobe_cns1_h_host () = __DATA:Adobe-CNS1-H-Host

let cmap_adobe_cns1_h_mac () = __DATA:Adobe-CNS1-H-Mac

let cmap_adobe_cns1_ucs2 () = __DATA:Adobe-CNS1-UCS2

let cmap_adobe_gb1_gbk_euc () = __DATA:Adobe-GB1-GBK-EUC

let cmap_adobe_gb1_gbpc_euc () = __DATA:Adobe-GB1-GBpc-EUC

let cmap_adobe_gb1_h_cid () = __DATA:Adobe-GB1-H-CID

let cmap_adobe_gb1_h_host () = __DATA:Adobe-GB1-H-Host

let cmap_adobe_gb1_h_mac () = __DATA:Adobe-GB1-H-Mac

let cmap_adobe_gb1_ucs2 () = __DATA:Adobe-GB1-UCS2

let cmap_adobe_japan1_90ms_rksj () = __DATA:Adobe-Japan1-90ms-RKSJ

let cmap_adobe_japan_90pv_rksj () = __DATA:Adobe-Japan1-90pv-RKSJ

let cmap_adobe_japan1_h_cid () = __DATA:Adobe-Japan1-H-CID

let cmap_adobe_japan1_h_host () = __DATA:Adobe-Japan1-H-Host

let cmap_adobe_japan1_h_mac () = __DATA:Adobe-Japan1-H-Mac

let cmap_adobe_japan1_ps_h () = __DATA:Adobe-Japan1-PS-H

let cmap_adobe_japan1_ps_v () = __DATA:Adobe-Japan1-PS-V

let cmap_adobe_japan1_ucs2 () = __DATA:Adobe-Japan1-UCS2

let cmap_adobe_korea1_h_cid () = __DATA:Adobe-Korea1-H-CID

let cmap_adobe_korea1_h_host () = __DATA:Adobe-Korea1-H-Host

let cmap_adobe_korea1_h_mac () = __DATA:Adobe-Korea1-H-Mac

let cmap_adobe_korea1_kscms_uhc () = __DATA:Adobe-Korea1-KSCms-UHC

let cmap_adobe_korea1_kscpc_euc () = __DATA:Adobe-Korea1-KSCpc-EUC

let cmap_adobe_korea1_ucs2 () = __DATA:Adobe-Korea1-UCS2

let cmap_adobe_kr_ucs2 () = __DATA:Adobe-KR-UCS2

let cmap_b5pc_ucs2 () = __DATA:B5pc-UCS2

let cmap_b5pc_ucs2c () = __DATA:B5pc-UCS2C

let cmap_eten_b5_ucs2 () = __DATA:ETen-B5-UCS2

let cmap_gbk_euc_ucs2 () = __DATA:GBK-EUC-UCS2

let cmap_gbpc_euc_ucs2 () = __DATA:GBpc-EUC-UCS2

let cmap_gbpc_euc_ucs2c () = __DATA:GBpc-EUC-UCS2C

let cmap_kscms_uhc_ucs2 () = __DATA:KSCms-UHC-UCS2

let cmap_kscpc_euc_ucs2 () = __DATA:KSCpc-EUC-UCS2

let cmap_kscpc_euc_ucs2c () = __DATA:KSCpc-EUC-UCS2C

let cmap_ucs2_90ms_rksj () = __DATA:UCS2-90ms-RKSJ

let cmap_ucs2_90pv_rksj () = __DATA:UCS2-90pv-RKSJ

let cmap_ucs2_b5pc () = __DATA:UCS2-B5pc

let cmap_ucs2_eten_b5 () = __DATA:UCS2-ETen-B5

let cmap_ucs2_gbk_euc () = __DATA:UCS2-GBK-EUC

let cmap_ucs2_gbpc_euc () = __DATA:UCS2-GBpc-EUC

let cmap_kscms_uhc () = __DATA:UCS2-KSCms-UHC

let cmap_ucs2_kscpc_euc () = __DATA:UCS2-KSCpc-EUC
