(* AFM data for standard 14 fonts *)

(* This file and the 14 PostScript(R) AFM files it accompanies may be used,
   copied, and distributed for any purpose and without charge, with or without
   modification, provided that all copyright notices are retained; that the AFM
   files are not distributed without this file; that all modifications to this
   file or any of the AFM files are prominently noted in the modified file(s);
   and that this paragraph is not modified. Adobe Systems has no responsibility
   or obligation to support the use of the AFM files. *)

let courier_afm () = __DATA:Courier.afm

let courier_oblique_afm () = __DATA:Courier-Oblique.afm

let courier_bold_afm () = __DATA:Courier-Bold.afm

let courier_bold_oblique_afm () = __DATA:Courier-BoldOblique.afm

let times_roman_afm () = __DATA:Times-Roman.afm

let times_bold_afm () = __DATA:Times-Bold.afm

let times_italic_afm () = __DATA:Times-Italic.afm

let times_bold_italic_afm () = __DATA:Times-BoldItalic.afm

let helvetica_afm () = __DATA:Helvetica.afm

let helvetica_bold_afm () = __DATA:Helvetica-Bold.afm

let helvetica_oblique_afm () = __DATA:Helvetica-Oblique.afm

let helvetica_bold_oblique_afm () = __DATA:Helvetica-BoldOblique.afm

let symbol_afm () = __DATA:Symbol.afm

let zapf_dingbats_afm () = __DATA:ZapfDingbats.afm
