(** Standard CMaps *)

val cmap_90ms_rksj_ucs2 : unit -> string

val cmap_90pv_rksj_ucs2 : unit -> string

val cmap_90pv_rksj_ucs2c : unit -> string

val cmap_adobe_cns1_b5pc : unit -> string

val cmap_adobe_cns1_eten_B5 : unit -> string

val cmap_adobe_cns1_h_cid : unit -> string

val cmap_adobe_cns1_h_host : unit -> string

val cmap_adobe_cns1_h_mac : unit -> string

val cmap_adobe_cns1_ucs2 : unit -> string

val cmap_adobe_gb1_gbk_euc : unit -> string

val cmap_adobe_gb1_gbpc_euc : unit -> string

val cmap_adobe_gb1_h_cid : unit -> string

val cmap_adobe_gb1_h_host : unit -> string

val cmap_adobe_gb1_h_mac : unit -> string

val cmap_adobe_gb1_ucs2 : unit -> string

val cmap_adobe_japan1_90ms_rksj : unit -> string

val cmap_adobe_japan_90pv_rksj : unit -> string

val cmap_adobe_japan1_h_cid : unit -> string

val cmap_adobe_japan1_h_host : unit -> string

val cmap_adobe_japan1_h_mac : unit -> string

val cmap_adobe_japan1_ps_h : unit -> string

val cmap_adobe_japan1_ps_v : unit -> string

val cmap_adobe_japan1_ucs2 : unit -> string

val cmap_adobe_korea1_h_cid : unit -> string

val cmap_adobe_korea1_h_host : unit -> string

val cmap_adobe_korea1_h_mac : unit -> string

val cmap_adobe_korea1_kscms_uhc : unit -> string

val cmap_adobe_korea1_kscpc_euc : unit -> string

val cmap_adobe_korea1_ucs2 : unit -> string

val cmap_adobe_kr_ucs2 : unit -> string

val cmap_b5pc_ucs2 : unit -> string

val cmap_b5pc_ucs2c : unit -> string

val cmap_eten_b5_ucs2 : unit -> string

val cmap_gbk_euc_ucs2 : unit -> string

val cmap_gbpc_euc_ucs2 : unit -> string

val cmap_gbpc_euc_ucs2c : unit -> string

val cmap_kscms_uhc_ucs2 : unit -> string

val cmap_kscpc_euc_ucs2 : unit -> string

val cmap_kscpc_euc_ucs2c : unit -> string

val cmap_ucs2_90ms_rksj : unit -> string

val cmap_ucs2_90pv_rksj : unit -> string

val cmap_ucs2_b5pc : unit -> string

val cmap_ucs2_eten_b5 : unit -> string

val cmap_ucs2_gbk_euc : unit -> string

val cmap_ucs2_gbpc_euc : unit -> string

val cmap_kscms_uhc : unit -> string

val cmap_ucs2_kscpc_euc : unit -> string
