(** AFM Data for the standard 14 fonts *)

val courier_afm : unit -> string

val courier_oblique_afm : unit -> string

val courier_bold_afm : unit -> string

val courier_bold_oblique_afm : unit -> string

val times_roman_afm : unit -> string

val times_bold_afm : unit -> string

val times_italic_afm : unit -> string

val times_bold_italic_afm : unit -> string

val helvetica_afm : unit -> string

val helvetica_bold_afm : unit -> string

val helvetica_oblique_afm : unit -> string

val helvetica_bold_oblique_afm : unit -> string

val symbol_afm : unit -> string

val zapf_dingbats_afm : unit -> string
(** We pass unit because the AFM files are parsed upon demand and then memoised. *)
